#version 450
#extension GL_ARB_separate_shader_objects : enable

layout(location = 0) in VS_OUT {
    vec3 normal;
    vec3 texCoords;
    vec3 worldPos;
} vs_out;

layout(location = 0) out vec4 outColor;

layout (std140, binding = 1) uniform DirectionalLights {
    vec4 colors[4];
    vec4 directions[4];
    int count;
} directionalLights;

layout (std140, binding = 0) uniform Camera {
    mat4 transformationProjectionMatrix;
    mat4 viewProjectionInverseMatrix;
    mat4 viewMatrix;
    mat4 projectionMatrix;
    ivec2 viewport;
    vec3 eyesPos;
} camera;

layout(binding = 2) uniform sampler2D surfaceTexture;

#define NUM_OCTAVES 8

// ---------------------------------------------------------------------------------------------------------------------
// Simplex 3D Noise
// by Ian McEwan, Ashima Arts
//
vec4 permute(vec4 x) {
    return mod(((x*34.0)+1.0)*x, 289.0);
}

vec4 taylorInvSqrt(vec4 r) {
    return 1.79284291400159 - 0.85373472095314 * r;
}

float snoise(vec3 v) {
    const vec2  C = vec2(1.0/6.0, 1.0/3.0);
    const vec4  D = vec4(0.0, 0.5, 1.0, 2.0);

    // First corner
    vec3 i  = floor(v + dot(v, C.yyy));
    vec3 x0 =   v - i + dot(i, C.xxx);

    // Other corners
    vec3 g = step(x0.yzx, x0.xyz);
    vec3 l = 1.0 - g;
    vec3 i1 = min(g.xyz, l.zxy);
    vec3 i2 = max(g.xyz, l.zxy);

    //  x0 = x0 - 0. + 0.0 * C
    vec3 x1 = x0 - i1 + 1.0 * C.xxx;
    vec3 x2 = x0 - i2 + 2.0 * C.xxx;
    vec3 x3 = x0 - 1. + 3.0 * C.xxx;

    // Permutations
    i = mod(i, 289.0);
    vec4 p = permute(permute(permute(
    i.z + vec4(0.0, i1.z, i2.z, 1.0))
    + i.y + vec4(0.0, i1.y, i2.y, 1.0))
    + i.x + vec4(0.0, i1.x, i2.x, 1.0));

    // Gradients
    // ( N*N points uniformly over a square, mapped onto an octahedron.)
    float n_ = 1.0/7.0;// N=7
    vec3  ns = n_ * D.wyz - D.xzx;

    vec4 j = p - 49.0 * floor(p * ns.z *ns.z);//  mod(p,N*N)

    vec4 x_ = floor(j * ns.z);
    vec4 y_ = floor(j - 7.0 * x_);// mod(j,N)

    vec4 x = x_ *ns.x + ns.yyyy;
    vec4 y = y_ *ns.x + ns.yyyy;
    vec4 h = 1.0 - abs(x) - abs(y);

    vec4 b0 = vec4(x.xy, y.xy);
    vec4 b1 = vec4(x.zw, y.zw);

    vec4 s0 = floor(b0)*2.0 + 1.0;
    vec4 s1 = floor(b1)*2.0 + 1.0;
    vec4 sh = -step(h, vec4(0.0));

    vec4 a0 = b0.xzyw + s0.xzyw*sh.xxyy;
    vec4 a1 = b1.xzyw + s1.xzyw*sh.zzww;

    vec3 p0 = vec3(a0.xy, h.x);
    vec3 p1 = vec3(a0.zw, h.y);
    vec3 p2 = vec3(a1.xy, h.z);
    vec3 p3 = vec3(a1.zw, h.w);

    //Normalise gradients
    vec4 norm = taylorInvSqrt(vec4(dot(p0, p0), dot(p1, p1), dot(p2, p2), dot(p3, p3)));
    p0 *= norm.x;
    p1 *= norm.y;
    p2 *= norm.z;
    p3 *= norm.w;

    // Mix final noise value
    vec4 m = max(0.6 - vec4(dot(x0, x0), dot(x1, x1), dot(x2, x2), dot(x3, x3)), 0.0);
    m = m * m;
    return 42.0 * dot(m*m, vec4(dot(p0, x0), dot(p1, x1),
    dot(p2, x2), dot(p3, x3)));
}

float fbm(vec3 x) {
    float v = 0.0;
    float a = 0.5;
    vec3 shift = vec3(100);
    for (int i = 0; i < NUM_OCTAVES; ++i) {
        v += a * snoise(x);
        x = x * 2.0 + shift;
        a *= 0.5;
    }
    return v;
}

const vec3 sunPos = vec3(0.0, 0.0, 0.0);

// ---------------------------------------------------------------------------------------------------------------------
vec4 surfaceColor() {
    float height = clamp(fbm(vs_out.texCoords * 1.35) + 0.5, 0.0, 1.0);

    vec3 surface = texture(surfaceTexture, vec2(height, 1.0)).rgb;

    vec3 light = normalize(sunPos - vs_out.worldPos);
    vec3 radiance = directionalLights.colors[0].xyz;

    float d = clamp(dot(vs_out.normal, light), 0.0, 1.0);
    vec3 ambient = vec3(d);

    return vec4(surface * ambient, 1.0);
}

// ---------------------------------------------------------------------------------------------------------------------
vec4 atmosphereColor() {
    vec4 atmosphere = vec4(0.0);

    vec4 atmosphereColorStart = vec4(0.1, 0.7, 1.0, 1.0);
    vec4 atmosphereColorEnd = vec4(0.8, 0.9, 1.0, 1.0);

    // Light scattering
    vec3 light = normalize(sunPos - vs_out.worldPos);
    vec3 radiance = directionalLights.colors[0].xyz;

    float d = clamp(dot(vs_out.normal, light), 0.0, 1.0);
    // d = pow(d, 1.0);

    vec3 eyesDir = normalize(camera.eyesPos - vs_out.worldPos);

    float f = 1.0 - clamp(dot(vs_out.normal, eyesDir), 0.0, 1.0);
    f = pow(f, 2.3);

    float factor = clamp(f * d, 0.0, 1.0);

    vec4 atmosphereColor = mix(atmosphereColorStart, atmosphereColorEnd, pow(f, 1.5));
    // atmosphereColor *= vec4(radiance, 1.0);

    return atmosphereColor * vec4(factor * 1.0);
}

// ---------------------------------------------------------------------------------------------------------------------
void main() {
    vec4 color = surfaceColor() + atmosphereColor();

    outColor = vec4(pow(color.rgb, vec3(2.2)), 1.0);
}
